<?php

namespace Drupal\entity_info_table\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\FieldableEntityInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Response;

/**
 *
 */
class EntityInfoTableController extends ControllerBase {

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
          $container->get('devel.dumper')
      );
  }

  /**
   * Generates and downloads a CSV file of the entity info table.
   */
  public function downloadCsv() {
    $build = $this->render();
    $headers = $build['entities']['#header'];
    $rows = $build['entities']['#rows'];
    $rows = array_map(function ($row) {
      return array_map(function ($value) {
        return $value['data'];
      }, $row);
    }, $rows);

    // Build CSV content using fputcsv
    ob_start();
    $handle = fopen('php://output', 'w');

    fputcsv($handle, $headers);
    foreach ($rows as $row) {
      fputcsv($handle, array_map(function($value) {
        return is_array($value) ? json_encode($value, JSON_UNESCAPED_UNICODE) : $value;
      }, array_values($row)));
    }
    fclose($handle);
    $output = ob_get_clean();

    // Set response headers for CSV download
    $response = new Response($output);
    $response->headers->set('Content-Type', 'text/csv');
    $response->headers->set('Content-Disposition', 'attachment; filename="entity_info_table.csv"');
    return $response;
  }

  /**
   *
   */
  public function render(): array {
    $build = [];
    $build['download_button'] = [
      '#type' => 'link',
      '#title' => $this->t('Download as CSV'),
      '#url' => \Drupal\Core\Url::fromRoute('entity_info_table.download_csv'),
      '#attributes' => [
        'class' => ['button', 'button--primary'],
      ],
    ];
    $build['download_button']['#weight'] = -10;

    $headers = [
      $this->t('Entity Type ID'),
      $this->t('Entity Type Name'),
      $this->t('Bundle'),
      $this->t('Bundle label'),
      $this->t('Field name'),
      $this->t('Field label'),
      $this->t('Field type'),
      $this->t('Target type'),
      $this->t('Target bundles'),
      $this->t('Allow values'),
      $this->t('Field description'),
    ];

    $rows = [];

    foreach ($this->entityTypeManager()->getDefinitions() as $entity_type_id => $entity_type) {
      // Fail with an exception for non-fieldable entity types.
      $bundle_list = \Drupal::service('entity_type.bundle.info')->getBundleInfo($entity_type_id);
      
      // webform & webform_submission are different entity type :)
      if ($entity_type_id == 'webform') {
        continue;
      }
      if (!$entity_type->entityClassImplements(FieldableEntityInterface::class)) {
        continue;
      }
      foreach ($bundle_list as $bundle_key => $bundle_info) {
        foreach (\Drupal::service('entity_field.manager')->getFieldDefinitions($entity_type_id, $bundle_key) as $field_name => $field_definition) {
          if (!empty($field_definition->getTargetEntityTypeId())) {
            $bundleFields[$entity_type_id][$bundle_key][$field_name]['type'] = $field_definition->getType();
            $bundleFields[$entity_type_id][$bundle_key][$field_name]['label'] = $field_definition->getLabel();
            $row['id'] = [
              'data' => $entity_type->id(),
              'filter' => TRUE,
            ];
            $row['name'] = [
              'data' => $entity_type->getLabel(),
              'filter' => TRUE,
            ];
            $row['bundle'] = [
              'data' => $bundle_key,
              'filter' => TRUE,
            ];
            $row['bundle_label'] = [
              'data' => $bundle_info['label'],
              'filter' => FALSE,
            ];
            $row['field_name'] = [
              'data' => $field_definition->getName(),
              'filter' => TRUE,
            ];
            $row['field_label'] = [
              'data' => $field_definition->getLabel(),
              'filter' => TRUE,
            ];
            $field_type = $field_definition->getType();
            $target_type = '';
            $target_bundles = [];
            if ($field_definition->getType() == 'entity_reference') {
              $target_type = $field_definition->getSetting('target_type');
              $target_bundles = $field_definition->getSetting('handler_settings')['target_bundles'] ?? [];
            }
            $row['field_type'] = [
              'data' => $field_type,
              'filter' => TRUE,
            ];
            $row['target_type'] = [
              'data' => $target_type,
              'filter' => TRUE,
            ];
            $row['target_bundles'] = [
              'data' => implode(', ', $target_bundles),
              'filter' => TRUE,
            ];
            $field_values = [];
            if (in_array($field_definition->getType(), ['list_integer', 'list_string'])) {
              $values = $field_definition->getSetting('allowed_values') ?? [];
              if ($values) {
                foreach ($values as $key => $label) {
                  $field_values []= "$key:$label";
                }
              }
            }
            $row['field_values'] = [
              'data' => implode(', ', $field_values),
              'filter' => TRUE,
            ];
            $field_description = $field_definition->getFieldStorageDefinition()->getDescription() ? $field_definition->getFieldStorageDefinition()->getDescription() : NULL;
            $row['field_description'] = [
              'data' => $field_description,
              'filter' => TRUE,
            ];
            $rows[] = $row;
          }
        }
      }
    }
    ksort($rows);

    $build['entities'] = [
      '#type' => 'devel_table_filter',
      '#filter_label' => $this->t('Search'),
      '#filter_placeholder' => $this->t('Enter entity type id, provider or class'),
      '#filter_description' => $this->t('Enter a part of the entity type id, provider or class to filter by.'),
      '#header' => $headers,
      '#rows' => $rows,
      '#empty' => $this->t('No entity types found.'),
      '#sticky' => TRUE,
      '#attributes' => [
        'class' => ['devel-entity-type-list'],
      ],
    ];

    return $build;
  }

}
